// RW Bumpmaster
// ==============
// Imageboard Retrieval & Archiving Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/curator/bumpmaster/bumpmaster.hpp
// -Declarations of the baseline RW Bumpmaster sub-library

#pragma once

#include "board.hpp"
#include "catalog.hpp"
#include "page.hpp"
#include "post.hpp"
#include "site.hpp"
#include "thread.hpp"

#include "incls_all.hpp"

namespace rw::curator::bumpmaster {

//------------------------------------------------------------------------------
/** A simple data struct to help with thread's post's information
 */
struct Post_data {  // Lynxchan's JSON tag names:
  string   date;    // creation
  string   msg;     // message
  unsigned id;      // postId
};

//------------------------------------------------------------------------------
/** Assign to a post's data struct from a JSON object
 * @param[in] j
 * @param[out] d
 */
void from_json(Json const& j, Post_data& p);

//------------------------------------------------------------------------------
/** Assign to a JSON object from a post's data struct
 * @param[out] j
 * @param[in] d
 */
void to_json(Json& j, Post_data const& p);

//------------------------------------------------------------------------------
/** A simple data struct to help with catalog's thread's information
 */
struct Thrd_data {  // Lynxchan's JSON tag names:
  string   msg;     // message
  string   subj;    // subject
  unsigned id;      // threadId

  vector<Post_data> posts;
};

//------------------------------------------------------------------------------
/** Assign to a thread's data struct from a JSON object
 * @param j
 * @param t
 */
void from_json(Json const& j, Thrd_data& t);

//------------------------------------------------------------------------------
/** Assign to a JSON object from a thread's data struct
 * @param j
 * @param t
 */
void to_json(Json& j, Thrd_data const& t);

//------------------------------------------------------------------------------
/**
 */
class Bumpmaster_b : public RW {
 public:
  Bumpmaster_b();                                    // default ctor
  explicit Bumpmaster_b(std::string const& handle);  // tag-parm'd ctor

  /** Extract site name, board name, from IB URI @a uri
   * @param cat_uri
   * @return
   */
  auto parse_ib_uri(std::string const& uri)
      -> std::tuple<std::string, std::string>;

  /**
   * @param site_nm
   * @param board_nm
   */
  // void chk_board_dir(std::string const& site_nm, std::string const&
  // board_nm);

  /** Assemble an IB catalog JSON URI from key parts
   * @param host
   * @param path
   * @return the assembled URI
   * @note: example resulting form: https://alogs.space/robowaifu/catalog.json
   */
  auto bld_cat_uri(std::string const& host, std::string const& path)
      -> std::string;

  /** Assemble an IB thread JSON URI from key parts
   * @param host
   * @param path
   * @param thrd_id
   * @return the assembled URI
   * @note: example resulting form: https://alogs.space/robowaifu/res/14409.json
   */
  auto bld_thrd_uri(std::string const& host,
                    std::string const& path,
                    unsigned const     thrd_id) -> std::string;

  /**
   * @param cat_json
   * @return
   */
  auto bump_thrd_dats(Json const& cat_json) -> std::vector<Thrd_data>;

  /**
   * @param cat_json
   * @param site
   * @param board
   * @return
   */
  auto bump_thrd_uris(Json const&        cat_json,
                      std::string const& site,
                      std::string const& board) -> std::vector<std::string>;

  /**
   * @param thrd_json
   * @return
   */
  auto thrd_posts(Json const& thrd_json) -> std::vector<Post_data>;
};

//------------------------------------------------------------------------------
/**
 */
class Bumpmaster : public Bumpmaster_b {
 public:
  Bumpmaster();                                    // default ctor
  explicit Bumpmaster(std::string const& handle);  // tag-parm'd ctor

  inline auto board() const -> Board& { return board_; }
  inline auto catalog() const -> Catalog& { return catalog_; }
  inline auto page() const -> Page& { return page_; }
  inline auto post() const -> Post& { return post_; }
  inline auto site() const -> Site& { return site_; }
  inline auto thread() const -> Thread& { return thread_; }

  inline auto waifusearch() const -> Waifusearch& { return waifusearch_; }

 private:
  /**
   */
  void preflight();

  unique_ptr<Board_b>   _board_{make_unique<Board_b>(t(), "bumpmaster_board")};
  unique_ptr<Catalog_b> _catalog_{
      make_unique<Catalog_b>(t(), "bumpmaster_catalog")};
  unique_ptr<Page_b>   _page_{make_unique<Page_b>(t(), "bumpmaster_page")};
  unique_ptr<Post_b>   _post_{make_unique<Post_b>(t(), "bumpmaster_post")};
  unique_ptr<Site_b>   _site_{make_unique<Site_b>(t(), "bumpmaster_site")};
  unique_ptr<Thread_b> _thread_{
      make_unique<Thread_b>(t(), "bumpmaster_thread")};

  unique_ptr<Waifusearch_b> _waifusearch_{
      make_unique<Waifusearch_b>("bumpmaster_waifusearch")};

  Board&   board_{reference_to<Board>(_board_.get())};
  Catalog& catalog_{reference_to<Catalog>(_catalog_.get())};
  Page&    page_{reference_to<Page>(_page_.get())};
  Post&    post_{reference_to<Post>(_post_.get())};
  Site&    site_{reference_to<Site>(_site_.get())};
  Thread&  thread_{reference_to<Thread>(_thread_.get())};

  Waifusearch& waifusearch_{reference_to<Waifusearch>(_waifusearch_.get())};

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

 public:
  /** Pull a JSON dolldrop
   * @param json_uri
   * @return true + @c Json object constructed from the dolldrop, if valid
   */
  auto pull_json_drop(std::string const& json_uri) -> std::tuple<bool, Json>;

  /** Pull a JSON dolldrop, with one retry
   * @param json_uri
   * @return true + @c Json object constructed from the dolldrop, if valid
   * @note: removes the old dolldrop from disk, upon failure
   * @note: primarily a wrapper mechanism for @c pull_json_drop()
   */
  auto pull_json_retry(std::string const& json_uri) -> std::tuple<bool, Json>;

  /** Load a container with (ostensibly) the Page 1 threads
   * @param cat_json
   * @param thrd_lim
   * @param site
   * @param board
   * @return
   * @note: merges thread's data from catalog, with post's data from thread
   */
  auto top_bump_thrds(Json const&        cat_json,
                      int const          thrd_lim,
                      std::string const& site,
                      std::string const& board) -> std::vector<Thrd_data>;

  /** Confirm the dolldrop file size on disk matches the stated size from header
   * @param drop_uri
   * @return true if sizes match
   */
  auto sz_match(std::string const& drop_uri) -> bool;

  /** Switchable signal to the @c Dolldrop member this is a catalog pull request
   * @param flag; default==true
   * @note: (indirectly) affects the @c Dolldrop_b::chkd_pull_g() algorithm
   */
  void do_catalog(bool const flag = true);

  inline auto dolldrop() const -> Dolldrop& { return dolldrop_; }

 private:
  unique_ptr<Dolldrop_b> _dolldrop_{
      make_unique<Dolldrop_b>(t(), "bumpmaster_dolldrop")};

  Dolldrop& dolldrop_{reference_to<Dolldrop>(_dolldrop_.get())};

#endif  // __has_include(<curl_easy.h>)
};

}  // namespace rw::curator::bumpmaster

//    --=^=--

// "______"

// rw_curator_lib.v220312
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
