//   RW Dollnet
// ==============
// External, Inter-Networking Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/dollnet_util.cpp
// -Definitions of some baseline RW Dollnet utility functions

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "../../../include/dollhouse/net/dollnet_util.hpp"

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------

Time start;  // will start the instant before an xfer attempt

//------------------------------------------------------------------------------
auto pct(float const xfer_frac) -> std::string {
  auto          pct{xfer_frac * 100};
  ostringstream pctss;
  pctss << setw(5) << setfill(' ') << fixed << setprecision(1) << pct;

  return pctss.str();
}

//------------------------------------------------------------------------------
auto ticks(float const xfer_frac, bool const do_rev) -> std::string {
  // part already done (the tickmarks)
  auto done = static_cast<int>(round(tot_ticks * xfer_frac));

  // part still remaining (the spaces)
  auto rem{tot_ticks - done};

  ostringstream tickss;
  if (do_rev)
    tickss << setw(rem) << setfill(' ') << "" << setw(done) << setfill('=')
           << "";
  else
    tickss << setw(done) << setfill('=') << "" << setw(rem) << setfill(' ')
           << "";

  return tickss.str();
}

//------------------------------------------------------------------------------
auto m_of_n(curl_off_t const xfer_tot,
            curl_off_t const xfer_done,
            float const      units) -> std::string {
  ostringstream m_of_nss;
  m_of_nss.imbue(locale{""});  // provides for digit separators
  m_of_nss << setw(5) << fixed << setprecision(1) << (xfer_done / units)
           << " of " << setw(5) << fixed << setprecision(1)
           << (xfer_tot / units);

  return m_of_nss.str();

  // https://en.cppreference.com/w/cpp/locale/locale
  // https://en.cppreference.com/w/cpp/io/basic_ios/imbue
}

//------------------------------------------------------------------------------
auto hms_fmt(Durn durn) -> std::string {  //

  auto const hrs = duration_cast<hours>(durn);
  durn -= hrs;
  auto const mins = duration_cast<minutes>(durn);
  durn -= mins;
  Secs_f const secs = durn;

  ostringstream hrss;
  ostringstream minss;

  auto const have_hrs{hrs.count() > 0};
  if (have_hrs)
    hrss << hrs.count() << 'h';

  auto const have_mins{mins.count() > 0};
  if (have_mins || have_hrs)
    minss << setw(2) << setfill(' ') << mins.count() << 'm';

  ostringstream hmss{};
  hmss << (have_hrs ? hrss.str() : "")                  //
       << ((have_mins || have_hrs) ? minss.str() : "")  //
       << setw(4) << setfill(' ') << fixed << setprecision(1) << secs.count()
       << 's';

  return hmss.str();
}

//------------------------------------------------------------------------------
auto hms() -> std::string {
  return hms_fmt(Call_clk::now() - start);
}

//------------------------------------------------------------------------------
void prt_bar(curl_off_t const xfer_tot,
             curl_off_t const xfer_cmplt,
             Dir const        dir,
             bool const       do_rev) {
  // cast prevents integer division
  auto xfer_frac = xfer_cmplt / static_cast<float>(xfer_tot);

  auto xfer_pct    = pct(xfer_frac);                     // percent string
  auto xfer_ticks  = ticks(xfer_frac, do_rev);           // tickmarks string
  auto xfer_m_of_n = m_of_n(xfer_tot, xfer_cmplt, KiB);  // bytes string

  cout << ((dir == Dir::up) ? "  u/l " : "  d/l ") << xfer_pct << "% ["
       << xfer_ticks << "]  KiB : " << xfer_m_of_n << "  " << hms();
}

//------------------------------------------------------------------------------
void prt_bytes(curl_off_t const xfer_done, Dir const dir) {
  ostringstream raw_bytess{};
  raw_bytess.imbue(locale{""});  // provides for digit separators
  raw_bytess << ((dir == Dir::up) ? "  u/l" : "  d/l") << "   MiB : " << setw(7)
             << setfill(' ') << fixed << setprecision(2) << (xfer_done / MiB)
             << "   " << hms();

  cout << raw_bytess.str();
}

//------------------------------------------------------------------------------
auto prt_prog(Address,
              curl_off_t dl_tot,
              curl_off_t dl_rcvd,
              curl_off_t ul_tot,
              curl_off_t ul_sent) -> int {  //

  bool have_ul{false}, have_dl{false};

  // UPLOAD
  if (ul_tot > 0) {  // have an upload to display?

    prt_bar(ul_tot, ul_sent, Dir::up);
    have_ul = true;
  } else if (ul_sent > 0) {  // unbounded ul going (streaming)?

    prt_bytes(ul_sent, Dir::up);
    have_ul = true;
  }
  if (have_ul)
    // line ending depends if we have a simultaneous dl (stacked bars, or not)
    cout << ((dl_tot > 0 || dl_rcvd > 0) ? "\n" : "\r");

  // DOWNLOAD
  if (dl_tot > 0) {  // have a download to display?

    prt_bar(dl_tot, dl_rcvd, Dir::dn, true);
    have_dl = true;
  } else if (dl_rcvd > 0) {  // unbounded dl going (streaming)?

    prt_bytes(dl_rcvd, Dir::dn);
    have_dl = true;
  }
  if (have_dl)
    // line ending depends if we have a simultaneous ul (upline + r, or just r)
    cout << ((ul_tot > 0 || ul_sent > 0) ? "\u008D\r" : "\r");

  // immediately show xfer times on curl requests, even w/o bytes flowing (yet)
  if (! (have_ul || have_dl))
    cout << "  req made" << setw(76) << setfill(' ') << hms() << '\r';

  // printing w/ flushing works smoother for this particular type case ('\r')
  cout << flush;

  // finally, xfer will abort w/o a '0' func ret here
  return 0;
}

//------------------------------------------------------------------------------
int prt_time(Address, curl_off_t, curl_off_t, curl_off_t, curl_off_t) {  //

  cout << "  req made" << setw(76) << setfill(' ') << hms() << '\r';
  cout << flush;

  return 0;
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "This apartment has ultra-high bandwidth illegal even to large corporations?"

// rw_dollhouse_lib.v220312
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
