//  RW Mailstop
// ==============
// ______
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/mailstop_util.cpp
// -Definitions of some baseline RW Mailstop utility functions

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "../../../include/dollhouse/net/mailstop_util.hpp"

// DESIGN: NOTE: RW Mailstop likely to be moved out of rw::net, over to rw::sys

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------

auto const droplog{dropdir / "dropmod.json"};  // dolldrops log's filename

static bool log_chkd{false};  // flag for checking once that the droplog exists

//------------------------------------------------------------------------------
void from_json(Json const& j, Dropdata& d) {  //

  if (! j["date"].is_null())
    d.date = j["date"];
  else
    d.date = "";

  if (! j["name"].is_null())
    d.name = j["name"];
  else
    d.name = "";

  if (! j["sz"].is_null())
    d.sz = j["sz"];
  else
    d.sz = "";

  if (! j["uri"].is_null())
    d.uri = j["uri"];
  else
    d.uri = "";
}

//------------------------------------------------------------------------------
void to_json(Json& j, Dropdata const& d) {
  j = Json{{"date", d.date}, {"name", d.name}, {"sz", d.sz}, {"uri", d.uri}};
}

//------------------------------------------------------------------------------
void mk_logfile() {
  ofstream ofs{droplog};
  ofs << "[]";  // insert a default empty JSON array
  ofs.close();
}

//------------------------------------------------------------------------------
void chk_logfile() {
  if (! exists_fs(droplog)) {  // confirm droplog exists
    mk_logfile();
  }

  log_chkd = true;
}

//------------------------------------------------------------------------------
void mk_tagfile(fs::path const& tag_fil) {
  ofstream ofs{tag_fil};
  ofs << "000\n";  // insert a default zero starting count
  ofs.close();
}

//------------------------------------------------------------------------------
auto filetag() -> std::string {
  auto const tag_fil{dropdir / "tagfile"};

  if (! exists_fs(tag_fil)) {  // confirm tagfile exists
    mk_tagfile(tag_fil);
  }

  // capture tagfile's contents into vars
  int    cnt{0};
  string ext{""};
  //
  ifstream ifs{tag_fil};
  ifs >> cnt >> ext;  // DESIGN: any validation req'd here ?
  ifs.close();

  // store incremented count back out to file for next dolldrop cycle
  ofstream ofs{tag_fil};
  ofs << setw(3) << setfill('0') << ++cnt << '\n'
      << ((ext != "") ? ext + '\n' : "");  // also write ext if present
  ofs.close();

  // create file tag from now-incremented tagfile data
  ostringstream tagss{};
  tagss << '_' << setw(3) << setfill('0') << cnt << ext;

  return tagss.str();
}

//------------------------------------------------------------------------------
auto bld_dropname() -> fs::path {
  string const afile = "anons_file" + filetag();
  return {dropdir / afile};
}

//------------------------------------------------------------------------------
auto chk_dropname(std::string const& uri) -> fs::path {
  fs::path res{""};

  if (auto prevfile = get_name(uri); ! prevfile.empty()) {  // reuse dolldrop
    res = prevfile;

  } else {  // create a new dolldrop
    res = bld_dropname();
  }

  return res;
}

//------------------------------------------------------------------------------
auto dropfield(std::stringstream& hdr_strm, std::string const& field_nm)
    -> std::string {  //

  auto lcfn{field_nm};  // 1st mk cp for lowercase'd searches
  to_lower(lcfn);       //  "

  for (string line; getline(hdr_strm, line);) {
    to_lower(line);

    if (line.find(lcfn) != str_end) {  // we have the correct header tag
      // rm the tag text itself
      auto fld_dat = line.substr(lcfn.size(), str_end);

      // rm '\r' char, if present in the data
      if (auto r_it = fld_dat.find('\r'); r_it != str_end)
        fld_dat.erase(r_it, 1);

      return fld_dat;  // just the facts ma'am  :^)
    }
  }

  return string{};  // dropfield not found
}

//------------------------------------------------------------------------------
auto find_log_rec(std::string const& uri, Json& json) -> Json::iterator {
  // have this uri's matching droplog record within the JSON container?
  return find_if(begin(json), end(json),
                 [&uri](Json const& j) { return uri == j["uri"]; });
}

//------------------------------------------------------------------------------
auto add_log_rec(Json&              json_out,
                 std::string const& uri,
                 std::string const& tag,
                 std::string const& tagdat) -> bool {
  Dropdata d{};

  if (tag == "date")
    d.date = tagdat;
  else if (tag == "name")
    d.name = tagdat;
  else if (tag == "sz")
    d.sz = tagdat;

  // one source param or other, always set the uri tag for a new record
  if (tag == "uri")
    d.uri = tagdat;
  else
    d.uri = uri;

  // constructs a new JSON item in-place, via to_json()
  return try_emplace_back(json_out, d);
}

//------------------------------------------------------------------------------
auto set_tag(std::string const& uri,
             std::string const& tag,
             std::string const& tagdat) -> bool {
  bool res{false};

  if (! log_chkd)
    chk_logfile();

  auto j_out = rd_json_file(droplog);

  if (auto rec_it = find_log_rec(uri, j_out); rec_it != end(j_out)) {  //

    // first get tag's ref, then copy assign tag data to it (code within parens)
    if (auto ref_chk = (try_idx_tag(rec_it, tag) = tagdat);
        ref_chk == tagdat)  // now check ref; was assignment a success?
      res = true;

  } else {
    res = add_log_rec(j_out, uri, tag, tagdat);
  }

  res &= wrt_json_file(j_out, droplog);  // if both rec update AND file write
  return res;
}
// wrappers-related
auto set_date(std::string const& uri, std::string const& datedat) -> bool {
  return set_tag(uri, "date", datedat);
}
auto set_name(std::string const& uri, std::string const& namedat) -> bool {
  return set_tag(uri, "name", namedat);
}
auto set_sz(std::string const& uri, std::string const& szdat) -> bool {
  return set_tag(uri, "sz", szdat);
}
auto set_uri(std::string const& uri, std::string const& uridat) -> bool {
  return set_tag(uri, "uri", uridat);
}

//------------------------------------------------------------------------------
auto get_tag(std::string const& uri, std::string const& tag) -> std::string {
  string res{""};

  if (! log_chkd)
    chk_logfile();

  auto j_in = rd_json_file(droplog);

  if (auto const rec_it = find_log_rec(uri, j_in); rec_it != end(j_in)) {  //

    res = try_idx_tag(rec_it, tag);  /////

  } else {  // uri's record not found; this call will add a new one (indirectly)
    set_uri(uri, uri);
  }

  return res;
}
// wrappers-related
auto get_date(std::string const& uri) -> std::string {
  return get_tag(uri, "date");
}
auto get_name(std::string const& uri) -> std::string {
  return get_tag(uri, "name");
}
auto get_sz(std::string const& uri) -> std::string {
  return get_tag(uri, "sz");
}

//------------------------------------------------------------------------------
auto chk_droptimes(std::string const& uri, std::stringstream& hdr)
    -> std::tuple<bool, std::string> {
  bool have_new{false};
  hdr.seekg(0);  // rewind stream

  auto const prevdate = get_date(uri);
  auto const currdate = dropfield(hdr, "last-modified: ");

  if (prevdate != currdate) {
    have_new = true;
    set_date(uri, currdate);
  }

  auto const name = get_name(uri);
  return make_tuple(have_new, name);
}

//------------------------------------------------------------------------------
auto chk_dropsizes(std::string const& uri, std::stringstream& hdr) -> bool {
  bool size_match{true};
  hdr.seekg(0);  // rewind stream

  auto const prevsize = get_sz(uri);
  auto const currsize = dropfield(hdr, "content-length: ");

  if (prevsize != currsize) {
    size_match = false;
    set_sz(uri, currsize);
  }

  return size_match;
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dollhouse_lib.v220312
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
